/*
 * Copyright (c) 2007-2009,2012 Anthony Mallet                  --  Tue Sep  4 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include "opencv2/core/core_c.h"
# include "opencv2/core/version.hpp"
# include "opencv2/imgproc/imgproc_c.h"
# include "opencv2/calib3d/calib3d.hpp"
#endif


/* --- local data ---------------------------------------------------- */

/** calibration works with at least that number of images */
#define CALIBRATION_MIN_IMAGES	3


typedef struct viam_bdata_t {
#ifdef VIAM_OPENCV_SUPPORT
  int rows, cols;	/**< size of the calibration pattern */
  double size;		/**< dimension of the chessboard squares */
#endif
} viam_bdata_t;

typedef struct viam_cdata_t {
#ifdef VIAM_OPENCV_SUPPORT
  int nimages;		/**< number of images */
  int pcount;		/**< number of image points */
  CvPoint2D32f *points;	/**< image points */
  CvPoint2D32f *prevcb;	/**< previous chessboard points */
  CvPoint3D64f *trans;	/**< extrinsic translation */
  CvPoint3D64f *rot;	/**< extrinsic rotation */
  CvPoint3D64f *pose;	/**< chessboard position */
  int nslots;		/**< number of slots in arrays above */
#endif /* VIAM_OPENCV_SUPPORT */
} viam_cdata_t;


struct viam_bankcalib_t {
  viam_bankcalibration_t b;	/**< calibration data */
  viam_bdata_t *data;		/**< temporary data while calibrating */
};

struct viam_cameracalib_t {
  viam_cameracalibration_t c;	/**< calibration data */
  viam_cdata_t *data;		/**< temporary data while calibrating */
};

#ifdef VIAM_OPENCV_SUPPORT
static void	viam_calibrate_mono(viam_handle_t vh, int nimages,
			CvSize imagesize, CvPoint2D32f *ipoints, int rows,
			int cols, double size, viam_distortionmodel_t dmodel,
			double intrinsic[9], double distortion[5],
			CvPoint3D64f *rotation, CvPoint3D64f *translation,
			double *result);
static void	viam_calibrate_stereo(viam_handle_t vh, viam_bank_t bank,
			double *precision);
static void	viam_getrectification(double P1[12], double P2[12], int imin,
			int jmin, int imax, int jmax, double *baseline,
			double *pbaseline, double W1[9], double W2[9],
			double R1[9], double R2[9], double intrinsinc[9]);
static void	viam_get_CAHV(const double P[12], double C[3], double A[3],
			double H[3], double V[3]);
static void	viam_get_AGT(const double A1[3], const double H1[3],
			const double V1[3], double G[3], double T[3]);
static void	viam_get_rectAGT(const double C1[3], const double A1[3],
			const double H1[3], const double V1[3],
			const double C2[3], const double A2[3],
			const double H2[3], const double V2[3],
			double A[3], double G[3], double T[3]);
static void	viam_get_rectcoeff(const double C1[3], const double A1[3],
			const double H1[3], const double V1[3],
			const double C2[3], const double A2[3],
			const double H2[3], const double V2[3], double A[3],
			double G[3], double T[3], int umin, int vmin,
			int umax, int vmax, double *u0, double *ku,
			double *v0, double *kv);
static void	viam_get_K(int u, int v, const double Vi[3],
			const double Hi[3], const double Ai[3],
			const double T[3], const double G[3],
			const double A[3], double *Kx, double *Ky);
static void	viam_warping_coeffs(double A1[3], double H1[3], double V1[3],
			double A[3], double H[3], double V[3], double W[9]);
static void	fillquad(IplImage *img, CvPoint2D32f v[4]);
static void	undistort(CvPoint2D32f o, CvPoint2D32f *u, const double a[9],
			const double coeff[4]);
#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_calibration_start ---------------------------------------- */

#ifdef VIAM_OPENCV_SUPPORT

/** \brief Start bank calibration procedure.
 *
 * The calibration procedure expect to have images with a *planar*
 * chessboard made of black and white *squares*. The chessboard can have
 * any number of rows and columns.
 *
 * \param[in] vh	Library handle.
 * \param[in] bank	Bank to be calibrated.
 * \param[in] type	Calibration type. Stereo works only with banks
 *			configured with two cameras.
 * \param[in] rows	Number of rows of the chessboard pattern.
 * \param[in] cols	Number of columns of the chessboard pattern.
 * \param[in] size	Dimension in meter of the squares of the
 *			chessboard.
 */

int
viam_calibration_start(viam_handle_t vh, viam_bank_t bank,
		       viam_caltype_t type, int rows, int cols, double size)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;
  int i, s;

  /* sanity checks */
  if (type == VIAM_CAL_STEREO && viam_bank_nimages(bank) != 2)
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);

  /* set up calibration data for the bank */
  bc = viam_bank_caldata(bank);
  if (!bc) {
    bc = malloc(sizeof(*bc));
    if (!bc) return viam_seterrno(vh, Eviamlib_MEMORY);
    viam_bank_setcaldata(bank, bc);
  } else
    if (bc->data) return viam_seterrno(vh, Eviamlib_CALIBRATING);

  bc->data = malloc(sizeof(*bc->data));
  if (!bc->data) return viam_seterrno(vh, Eviamlib_MEMORY);

  bc->b.type = type;
  bc->data->rows = rows - 1;
  bc->data->cols = cols - 1;
  bc->data->size = size;

  /* set up calibration data for all cameras */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));
    if (!cc) {
      cc = malloc(sizeof(*cc));
      if (!cc) { s = Eviamlib_MEMORY; goto error; }
      viam_image_setcaldata(viam_bank_image(vh, bank, i), cc);
    } else
      if (cc->data) { s = Eviamlib_CALIBRATING; goto error; }

    cc->data = malloc(sizeof(*cc->data));
    if (!cc->data) { s = Eviamlib_MEMORY; goto error; }

    cc->data->nimages = 0;
    cc->data->pcount = 0;
    cc->data->points = NULL;
    cc->data->prevcb =
      malloc(bc->data->rows*bc->data->cols*sizeof(*cc->data->prevcb));
    if (!cc->data->prevcb) { s = Eviamlib_MEMORY; goto error; }
    cc->data->trans = NULL;
    cc->data->rot = NULL;
    cc->data->pose = NULL;
    cc->data->nslots = 0;
  }

  return 0;
error:
  if (bc->data) { free(bc->data); bc->data = NULL; }

  for(i=0; i<viam_bank_nimages(bank); i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));
    if (cc && cc->data) { free(cc->data); cc->data = NULL; }
  }
  return viam_seterrno(vh, s);
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_calibration_addimage ------------------------------------- */

#ifdef VIAM_OPENCV_SUPPORT

/** \brief Add calibration images for the bank.
 *
 * The calibration procedure expect to have images with a *planar*
 * chessboard made of black and white *squares*. The chessboard can have
 * any number of rows and columns.
 *
 * \param[in] vh	Library handle.
 * \param[in] bank	Bank being calibrated.
 * \param[in] image	User-supplied image array, with allocated buffers.
 * \param[in] nimages	Number of images in image array.
 * \param[in] overlapth	Minimal distance to previous chessboards (pixels).
 * \param[in] motionth	Maximal motion of the chessboard for validation (pixels).
 * \param[in] feedback	Display calibration pattern.
 * \param[out] result	True if chessboard was successfully found in all
 *			images.
 */

int
viam_calibration_addimage(viam_handle_t vh, viam_bank_t bank,
			  IplImage *image[], int nimages, double overlapth,
			  double motionth, viam_onoff_t feedback, int *result)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;
  CvPoint2D32f *p;
  CvPoint3D64f *t;
  CvPoint2D32f dir, min, max;
  int i, j, s, count;

  if (nimages != viam_bank_nimages(bank))
    return viam_seterrno(vh, Eviamlib_BAD_IMAGE_COUNT);

  bc = viam_bank_caldata(bank);
  if (!bc || !bc->data)
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  /* iterate over all all images of the bank */
  *result = 1;
  viam_seterrno(vh, 0);
  for(i=0; i<viam_bank_nimages(bank); i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));
    if (!cc || !cc->data)
      return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

    if (cc->data->pcount == 0) {
      cc->c.width = image[i]->width;
      cc->c.height = image[i]->height;
    } else if (image[i]->width != cc->c.width || image[i]->height != cc->c.height) {
      *result = 0;
      viam_seterrno(vh, Eviamlib_IMAGE_GEOMETRY);
      break;
    }

    /* resize temporary arrays that hold image points */
    if (cc->data->nslots <= cc->data->nimages) {
      p = realloc(cc->data->points,
		  (cc->data->pcount + bc->data->rows*bc->data->cols)*
		  sizeof(CvPoint2D32f));
      if (!p) return viam_seterrno(vh, Eviamlib_MEMORY);
      cc->data->points = p;

      t = realloc(cc->data->trans, (cc->data->nimages + 1) * sizeof(CvPoint3D64f));
      if (!t) return viam_seterrno(vh, Eviamlib_MEMORY);
      cc->data->trans = t;

      t = realloc(cc->data->rot, (cc->data->nimages + 1) * sizeof(CvPoint3D64f));
      if (!t) return viam_seterrno(vh, Eviamlib_MEMORY);
      cc->data->rot = t;

      t = realloc(cc->data->pose, (cc->data->nimages + 1) * sizeof(CvPoint3D64f));
      if (!t) return viam_seterrno(vh, Eviamlib_MEMORY);
      cc->data->pose = t;

      cc->data->nslots++;
    }

    /* look for chessboard points */
    p = cc->data->points + cc->data->pcount;
    s = cvFindChessboardCorners(
      image[i], cvSize(bc->data->cols, bc->data->rows),
      p, &count, CV_CALIB_CB_NORMALIZE_IMAGE | CV_CALIB_CB_ADAPTIVE_THRESH);

    if (s) {
      /* reject corners that are too close to the image edges */
      for(j=0; j<count; j++) {
	if (p[j].x < 1 || p[j].x > image[i]->width - 2 ||
	    p[j].y < 1 || p[j].y > image[i]->height - 2) {
	  s = 0; break;
	}
      }
    }

    if (s)
      cvFindCornerSubPix(image[i], p, count, cvSize(5, 5), cvSize(-1, -1),
			 cvTermCriteria(CV_TERMCRIT_EPS, 100, 0.001));

    if (feedback) {
      /* draw all previous rigs */
      CvPoint2D32f rig[4];
      p = cc->data->points;
      if (cc->data->nimages > 0)
      for(j=0; j < cc->data->nimages; j++, p+=bc->data->rows*bc->data->cols) {
	rig[0] = p[0];
	rig[1] = p[bc->data->cols-1];
	rig[2] = p[bc->data->cols*bc->data->rows-1];
	rig[3] = p[bc->data->cols*(bc->data->rows-1)];

	fillquad(image[i], rig);
      }

      /* and current points */
      cvDrawChessboardCorners(image[i], cvSize(bc->data->cols, bc->data->rows),
			      p, count, s);
    }
    if (!s) { *result = 0; continue; }

    /* check that chessboard did not move since the previous image */
    for(j=0;j<count; j++) {
      if (fabs(cc->data->prevcb[j].x - p[j].x) > motionth ||
	  fabs(cc->data->prevcb[j].y - p[j].y) > motionth)
	*result = 0;
      cc->data->prevcb[j] = p[j];
    }
    if (!*result) continue;

    /* estimate chessboard position by averaging corners position */
    cc->data->pose[cc->data->nimages] = cvPoint3D64f(0., 0., 0.);
    for(j=0; j<count; j++) {
      cc->data->pose[cc->data->nimages].x += p[j].x;
      cc->data->pose[cc->data->nimages].y += p[j].y;
    }
    cc->data->pose[cc->data->nimages].x /= count;
    cc->data->pose[cc->data->nimages].y /= count;

    /* estimate chessboard surface */
    min = max = p[0];
    for(j=1; j<count; j++) {
      if (min.x > p[j].x) min.x = p[j].x;
      if (min.y > p[j].y) min.y = p[j].y;
      if (max.x < p[j].x) max.x = p[j].x;
      if (max.y < p[j].y) max.y = p[j].y;
    }
    cc->data->pose[cc->data->nimages].z = (max.x - min.x)*(max.y - min.y);

    /* check that chessboard is at more than distth of other images */
    for(j=0; j<cc->data->nimages; j++)
      if (fabs(cc->data->pose[cc->data->nimages].z/cc->data->pose[j].z - 1.0) <
	  0.1 &&
	  hypotf(cc->data->pose[cc->data->nimages].x - cc->data->pose[j].x,
		 cc->data->pose[cc->data->nimages].y - cc->data->pose[j].y) <
	  overlapth) {
	  *result = 0; break;
      }
    if (!*result) continue;

    /* check the orientation is coherent with previous image */
    if (i == 0) {
      dir.x = p[count-1].x - p[0].x;
      dir.y = p[count-1].y - p[0].y;
    } else if (dir.x*(p[count-1].x - p[0].x) + dir.y*(p[count-1].y - p[0].y) < 0.) {
      *result = 0; continue;
    }
  }
#if 0
  if (cc->data->nimages >= CALIBRATION_MIN_IMAGES) for(s=0;s<2;s++) {
    unsigned char *im = malloc(image[s]->width*image[s]->height);

    memcpy(im, image[s]->imageData, image[s]->width*image[s]->height);
    memset(image[s]->imageData, 0, image[s]->width*image[s]->height);

    cc = viam_image_caldata(viam_bank_image(vh, bank, s));

    for(i=0;i<image[0]->width;i++) for(j=0;j<image[0]->height;j++) {
      CvPoint2D32f p, q;
      CvMat _i = cvMat(3, 3, CV_64FC1, cc->c.intrinsic);
      /* XXX reorder distortion */ CvMat _d = cvMat(1, 4, CV_64FC1, cc->c.distortion);
      double *w = cc->c.rectification;
      double k;

      q.x = w[0] * i + w[1] * j + w[2];
      q.y = w[3] * i + w[4] * j + w[5];
      k = w[6] * i + w[7] * j + w[8];
      q.x /= k;
      q.y /= k;
      distort(q, &p, &_i, &_d);

      if (p.x >= 0 && p.x < image[0]->width &&
	  p.y >= 0 && p.y < image[0]->height)
	((unsigned char *)image[s]->imageData)[i+j*image[0]->width] =
	  im[(int)p.x+(int)p.y*image[0]->width];
    }
    free(im);
  }
#endif
  if (!*result) return 0;

  /* remember all chessboard points for all images */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));

    if (feedback)
      printf("image %02d:%d, pose %f,%f, size %f\n", cc->data->nimages, i,
	     cc->data->pose[cc->data->nimages].x,
	     cc->data->pose[cc->data->nimages].y,
	     cc->data->pose[cc->data->nimages].z);

    cc->data->pcount += bc->data->rows * bc->data->cols;
    cc->data->nimages++;
  }

  return 0;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_calibration_estimate ------------------------------------- */

#ifdef VIAM_OPENCV_SUPPORT

/** \brief Compute calibration matrices.
 *
 * \param[in] vh		Library handle.
 * \param[in] bank		Bank being calibrated.
 * \param[out] precision	Back projection error.
 */

int
viam_calibration_estimate(viam_handle_t vh, viam_bank_t bank,
			  viam_distortionmodel_t dmodel, double *result)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;
  int i;

  bc = viam_bank_caldata(bank);
  if (!bc || !bc->data)
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  /* iterate over images */
  for(i=0; i<viam_bank_nimages(bank); i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));
    if (!cc || !cc->data)
      return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

    if (cc->data->nimages < CALIBRATION_MIN_IMAGES) {
      *result = (cc->c.width + cc->c.height)/2.;
      return 0;
    }

    viam_calibrate_mono(vh, cc->data->nimages, cvSize(cc->c.width, cc->c.height),
			cc->data->points,
			bc->data->rows, bc->data->cols, bc->data->size, dmodel,
			cc->c.intrinsic, cc->c.distortion,
			cc->data->rot, cc->data->trans, result);
  }

  /* estimate rectification for stereo banks */
  if (bc->b.type == VIAM_CAL_STEREO) {
    double precision;

    viam_calibrate_stereo(vh, bank, &precision);
  }

  return 0;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_calibration_finish --------------------------------------- */

#ifdef VIAM_OPENCV_SUPPORT

/** \brief Stop bank configuration procedure and compute calibration.
 *
 * \param[in] vh	Library handle.
 * \param[in] bank	Bank to be calibrated.
 */

int
viam_calibration_finish(viam_handle_t vh, viam_bank_t bank)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;
  int i;

  bc = viam_bank_caldata(bank);
  if (!bc || !bc->data)
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  /* clean-up calibration data for all cameras */
  viam_seterrno(vh, Eviamlib_OK);
  for(i=0; i<viam_bank_nimages(bank); i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));
    if (!cc || !cc->data) {
      viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);
      continue;
    }

    if (cc->data->points) free(cc->data->points);
    if (cc->data->trans) free(cc->data->trans);
    if (cc->data->rot) free(cc->data->rot);
    free(cc->data);
    cc->data = NULL;
  }

  free(bc->data);
  bc->data = NULL;
  return 0;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_image_calibration ---------------------------------------- */

int
viam_image_calibration(viam_handle_t vh, viam_image_ref image,
		       viam_cameracalibration_t *cal, const viam_geometry_t *geom)
{
  double sx, sy;
  IplImage *raw;
  viam_cameracalib_t cc = viam_image_caldata(image);
  if (!cc || !cal) return viam_seterrno(vh, Eviamlib_NOT_CALIBRATED);

  *cal = cc->c;
  if (!geom) return 0;

  /* get scaling factor */
  raw = viam_camera_raw(viam_image_camera(image));
  if (!raw || raw->width < 1 || raw->height < 1) return 0;

  sx = raw->width*geom->scalex / cal->width;
  sy = raw->height*geom->scaley / cal->height;

  /* translate according to geom */
  cal->intrinsic[2] -= geom->offx/sx;
  cal->intrinsic[5] -= geom->offy/sy;

  cal->intrirect[2] -= geom->offx/sx;
  cal->intrirect[5] -= geom->offy/sy;

  cal->rectification[2] -= geom->offx/sx;
  cal->rectification[5] -= geom->offy/sy;

  /* scale according to geom */
  cal->intrinsic[0] *= sx;
  cal->intrinsic[2] *= sx;
  cal->intrinsic[4] *= sy;
  cal->intrinsic[5] *= sy;

  cal->intrirect[0] *= sx;
  cal->intrirect[2] *= sx;
  cal->intrirect[4] *= sy;
  cal->intrirect[5] *= sy;

  cal->rectification[2] *= sx;
  cal->rectification[5] *= sy;
  cal->rectification[6] *= sx;
  cal->rectification[7] *= sy;

  cal->width *= sx;
  cal->height *= sy;

  return 0;
}


/* --- viam_bank_calibration ----------------------------------------- */

int
viam_bank_calibration(viam_handle_t vh, viam_bank_t bank,
		      viam_bankcalibration_t *cal)
{
  viam_bankcalib_t bc = viam_bank_caldata(bank);
  if (!bc || !cal) return viam_seterrno(vh, Eviamlib_NOT_CALIBRATED);

  *cal = bc->b;
  return 0;
}


/* --- viam_calibration_save ----------------------------------------- */

/** Save calibration data to file
 *
 * \param[in] vh	Library handle
 * \param[in] bank	Bank to save
 * \param[in] file	File name
 */

int
viam_calibration_save(viam_handle_t vh, viam_bank_t bank, const char *file)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;
  viam_image_ref c;
  time_t t;
  FILE *f;
  int i;

  bc = viam_bank_caldata(bank);
  if (!bc) return viam_seterrno(vh, Eviamlib_NOT_CALIBRATED);

  f = fopen(file, "w");
  if (!f) return viam_seterrno(vh, Eviamlib_FILE_IO);

  t = time(NULL);
  if (t != (time_t)-1) fprintf(f, "#\n# %s#\n", ctime(&t));

  fprintf(f, "\n# --- bank %s\n", viam_bank_name(bank));

  switch(bc->b.type) {
    case VIAM_CAL_MONO:
      fprintf(f, "monocular\n");
      break;

    case VIAM_CAL_STEREO:
      fprintf(f, "stereo\n");
      fprintf(f, "baseline\t%+.16e\n", bc->b.baseline);
      fprintf(f, "pixel baseline\t%+.16e\n", bc->b.pbaseline);
  }
  fprintf(f, "\n");

  for(i=0; i<viam_bank_nimages(bank); i++) {
    c = viam_bank_image(vh, bank, i);
    cc = viam_image_caldata(c);
    if (!cc) continue;

    fprintf(f, "\n# --- image %s\n", viam_image_name(c));

    fprintf(f, "camera %s:%s\n",
	    viam_hwdev_name(viam_camera_hwdev(viam_image_camera(c))),
	    viam_camera_uid(viam_image_camera(c)));
    fprintf(f, "size %dx%d\n", cc->c.width, cc->c.height);
    fprintf(f, "intrinsic\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n",
	    cc->c.intrinsic[0], cc->c.intrinsic[1], cc->c.intrinsic[2],
	    cc->c.intrinsic[3], cc->c.intrinsic[4], cc->c.intrinsic[5],
	    cc->c.intrinsic[6], cc->c.intrinsic[7], cc->c.intrinsic[8]);
    fprintf(f, "\n");
    fprintf(f, "intrinsic rectified\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n",
	    cc->c.intrirect[0], cc->c.intrirect[1], cc->c.intrirect[2],
	    cc->c.intrirect[3], cc->c.intrirect[4], cc->c.intrirect[5],
	    cc->c.intrirect[6], cc->c.intrirect[7], cc->c.intrirect[8]);
    fprintf(f, "\n");
    fprintf(f, "distortion\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\n",
	    cc->c.distortion[0], cc->c.distortion[1], cc->c.distortion[2],
	    cc->c.distortion[3], cc->c.distortion[4]);
    fprintf(f, "\n");
    fprintf(f, "rectification\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n",
	    cc->c.rectification[0], cc->c.rectification[1], cc->c.rectification[2],
	    cc->c.rectification[3], cc->c.rectification[4], cc->c.rectification[5],
	    cc->c.rectification[6], cc->c.rectification[7], cc->c.rectification[8]);
    fprintf(f, "\n");
    fprintf(f, "rotation\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n%+.16e\t%+.16e\t%+.16e\n",
	    cc->c.rotation[0], cc->c.rotation[1], cc->c.rotation[2],
	    cc->c.rotation[3], cc->c.rotation[4], cc->c.rotation[5],
	    cc->c.rotation[6], cc->c.rotation[7], cc->c.rotation[8]);

    fprintf(f, "\n");
  }

  fclose(f);
  return 0;
}


/* --- viam_calibration_load ----------------------------------------- */

/** Load calibration data from file
 *
 * \param[in] vh	Library handle
 * \param[in] bank	Bank to load
 * \param[in] file	File name
 */

int
viam_calibration_load(viam_handle_t vh, viam_bank_t bank, const char *file)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;
  viam_image_ref c;
  char line[256], *s;
  FILE *f;
  int i,j, l;

  bc = viam_bank_caldata(bank);
  if (!bc) {
    bc = malloc(sizeof(*bc));
    if (!bc) return viam_seterrno(vh, Eviamlib_MEMORY);
    viam_bank_setcaldata(bank, bc);

    bc->data = NULL;
  } else
    if (bc->data) return viam_seterrno(vh, Eviamlib_CALIBRATING);

  f = fopen(file, "r");
  if (!f) { viam_seterrno(vh, Eviamlib_FILE_IO); goto error; }
  l = 0;

#define getline(f, line)					\
  do {								\
    s = fgets(line, sizeof(line), f);				\
    if (!s) {viam_seterrno(vh, Eviamlib_FILE_IO); goto error;}	\
    ++l;							\
  } while(s[0] == '#' || s[0] == '\n')
#define scanerr(file, word)					\
  do {								\
    warnx("%s:%d: expecting %s keyword\n", file, l, word);	\
    viam_seterrno(vh, Eviamlib_FILE_IO);			\
    goto error;							\
  } while(0)

  getline(f, line);
  if (strstr(line, "monocular"))
    bc->b.type = VIAM_CAL_MONO;
  else if (strstr(line, "stereo"))
    bc->b.type = VIAM_CAL_STEREO;
  else scanerr(file, "monocular or stereo");

  switch(bc->b.type) {
    case VIAM_CAL_MONO:
      bc->b.baseline = 0.;
      bc->b.pbaseline = 0.;
      break;

    case VIAM_CAL_STEREO:
      getline(f, line);
      if (sscanf(line, "baseline %lg", &bc->b.baseline) != 1)
	scanerr(file, "baseline");
      getline(f, line);
      if (sscanf(line, "pixel baseline %lg", &bc->b.pbaseline) != 1)
	scanerr(file, "pixel baseline");
      break;
  }

  for(i=0; i<viam_bank_nimages(bank); i++) {
    c = viam_bank_image(vh, bank, i);
    cc = viam_image_caldata(c);
    if (!cc) {
      cc = malloc(sizeof(*cc));
      if (!cc) { viam_seterrno(vh, Eviamlib_MEMORY); goto error; }
      viam_image_setcaldata(c, cc);
      cc->data = NULL;
    } else
      if (cc->data) return viam_seterrno(vh, Eviamlib_CALIBRATING);

    getline(f, line);
    if (sscanf(line, "camera %*[^:]:%*s") != 0)
      scanerr(file, "camera uuid");

    getline(f, line);
    if (sscanf(line, "size %dx%d", &cc->c.width, &cc->c.height) != 2)
      scanerr(file, "size wxh");

    getline(f, line);
    if (!strstr(line, "intrinsic")) scanerr(file, "intrinsic");
    for(j=0; j<9; j+=3) {
      getline(f, line);
      if (sscanf(line, "%lg %lg %lg",
		 &cc->c.intrinsic[j], &cc->c.intrinsic[j+1], &cc->c.intrinsic[j+2]) != 3)
	scanerr(file, "3 doubles");
    }

    getline(f, line);
    if (!strstr(line, "intrinsic rectified")) scanerr(file, "intrinsic rectified");
    for(j=0; j<9; j+=3) {
      getline(f, line);
      if (sscanf(line, "%lg %lg %lg",
		 &cc->c.intrirect[j], &cc->c.intrirect[j+1], &cc->c.intrirect[j+2]) != 3)
	scanerr(file, "3 doubles");
    }

    getline(f, line);
    if (!strstr(line, "distortion")) scanerr(file, "distortion");
    getline(f, line);
    if (sscanf(line, "%lg %lg %lg",
	       &cc->c.distortion[0], &cc->c.distortion[1], &cc->c.distortion[2]) != 3)
      scanerr(file, "3 doubles");
    getline(f, line);
    if (sscanf(line, "%lg %lg",
	       &cc->c.distortion[3], &cc->c.distortion[4]) != 2)
      scanerr(file, "2 doubles");

    getline(f, line);
    if (!strstr(line, "rectification")) scanerr(file, "rectification");
    for(j=0; j<9; j+=3) {
      getline(f, line);
      if (sscanf(line, "%lg %lg %lg", &cc->c.rectification[j],
		 &cc->c.rectification[j+1], &cc->c.rectification[j+2]) != 3)
	scanerr(file, "3 doubles");
    }

    getline(f, line);
    if (!strstr(line, "rotation")) scanerr(file, "rotation");
    for(j=0; j<9; j+=3) {
      getline(f, line);
      if (sscanf(line, "%lg %lg %lg", &cc->c.rotation[j],
		 &cc->c.rotation[j+1], &cc->c.rotation[j+2]) != 3)
	scanerr(file, "3 doubles");
    }
  }

  fclose(f);
  return 0;

error:
  viam_bank_setcaldata(bank, NULL);
  free(bc);
  for(i=0; i<viam_bank_nimages(bank); i++) {
    c = viam_bank_image(vh, bank, i);
    cc = viam_image_caldata(c);
    if (cc) {
      viam_image_setcaldata(c, NULL);
      free(cc);
    }
  }
  return viam_errno(vh);
}


#ifdef VIAM_OPENCV_SUPPORT

/* --- viam_calibrate_mono ------------------------------------------- */

static void
viam_calibrate_mono(viam_handle_t vh, int nimages, CvSize imagesize,
		    CvPoint2D32f *ipoints, int rows, int cols, double size,
		    viam_distortionmodel_t dmodel, double intrinsic[9],
		    double distortion[5], CvPoint3D64f *rotation,
		    CvPoint3D64f *translation, double *result)
{
  CvMat _i = cvMat(3, 3, CV_64FC1, intrinsic);
  CvMat _d = cvMat(1, 5, CV_64FC1, distortion);
  CvMat *opointsmat, *pcount;
  CvMat  ipointsmat, tmat, rmat;
  double r;
  int flags;
  int i,j,k;

  /* allocate temporary matrices */
  opointsmat = cvCreateMat(nimages*rows*cols, 3, CV_64FC1);
  for(i=0; i<nimages; i++)
    for(j=0; j<rows; j++) for(k=0; k<cols; k++)
      ((CvPoint3D64f *)opointsmat->data.db)[(i*rows + j)*cols + k] =
	cvPoint3D64f(k * size, j * size, 0);

  pcount = cvCreateMat(nimages, 1, CV_32SC1);
  for(i=0; i<nimages; i++) pcount->data.i[i] = rows * cols;

  ipointsmat = cvMat(nimages*rows*cols, 2, CV_32FC1, ipoints);
  tmat = cvMat(nimages, 3, CV_64FC1, translation);
  rmat = cvMat(nimages, 3, CV_64FC1, rotation);

  /* calibrate */
  flags = 0;
  switch(dmodel) {
    case VIAM_DIST_R1T0:
    case VIAM_DIST_R1T2:
      distortion[1] = 0.;
      flags |= CV_CALIB_FIX_K2;
      /*FALLTHROUGH*/
    case VIAM_DIST_R2T0:
    case VIAM_DIST_R2T2:
      distortion[4] = 0.;
      flags |= CV_CALIB_FIX_K3;
      break;

    default: break;
  }
  switch(dmodel) {
    case VIAM_DIST_R1T0:
    case VIAM_DIST_R2T0:
    case VIAM_DIST_R3T0:
      distortion[2] = distortion[3] = 0.;
      flags |= CV_CALIB_ZERO_TANGENT_DIST;
      break;

    default: break;
  }

#if CV_MAJOR_VERSION==2 && CV_MINOR_VERSION<4
  cvCalibrateCamera2(opointsmat, &ipointsmat, pcount, imagesize,
                     &_i, &_d, &rmat, &tmat, flags);
#else
  cvCalibrateCamera2(opointsmat, &ipointsmat, pcount, imagesize,
                     &_i, &_d, &rmat, &tmat, flags,
                     cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS, 30,
                                    DBL_EPSILON));
#endif

  /* estimate reprojection error */
  *result = 0.;
  for(i=0; i<nimages; i++) {
    CvMat o = cvMat(rows*cols, 3, CV_64FC1,
		    &((CvPoint3D64f *)opointsmat->data.db)[i*rows*cols]);
    CvMat r = cvMat(1, 3, CV_64FC1, rotation + i);
    CvMat t = cvMat(1, 3, CV_64FC1, translation + i);
    CvMat *err = cvCreateMat(rows * cols, 2, CV_64FC1);

    cvProjectPoints2(&o, &r, &t, &_i, &_d, err,
		     NULL, NULL, NULL, NULL, NULL, 0.);

    for(j=0; j<rows; j++) for(k=0; k<cols; k++) {
      *result += hypot(
	((CvPoint2D32f *)ipointsmat.data.fl)[(i*rows + j)*cols + k].x -
	((CvPoint2D64f *)err->data.db)[j*cols + k].x,

	((CvPoint2D32f *)ipointsmat.data.fl)[(i*rows + j)*cols + k].y -
	((CvPoint2D64f *)err->data.db)[j*cols + k].y);
    }

    cvReleaseMat(&err);
  }
  *result /= nimages * rows * cols;

  /* reorder distortion coefficients */
  r = distortion[4];
  distortion[4] = distortion[3];
  distortion[3] = distortion[2];
  distortion[2] = r;

  cvReleaseMat(&opointsmat);
  cvReleaseMat(&pcount);
}


/* --- viam_calibrate_stereo ----------------------------------------- */

static void
viam_calibrate_stereo(viam_handle_t vh, viam_bank_t bank, double *precision)
{
  viam_bankcalib_t bc;
  viam_cameracalib_t cc;

  int nimages, i, j, k;
  double base, pbase, err;
  double rot[2][9], ir[9], e[12];
  double p[2][12], w[2][9], wi[2][9];

  CvMat _i = cvMat(3, 3, CV_64FC1, NULL);
  CvMat rvec = cvMat(1, 3, CV_64FC1, NULL);
  CvMat emat = cvMat(3, 4, CV_64FC1, e);
  CvMat pmat[2] = { cvMat(3, 4, CV_64FC1, p[0]), cvMat(3, 4, CV_64FC1, p[1]) };
  CvMat wmat[2] = { cvMat(3, 3, CV_64FC1, w[0]), cvMat(3, 3, CV_64FC1, w[1]) };
  CvMat rotmat[2] = { cvMat(3, 3, CV_64FC1, rot[0]), cvMat(3, 3, CV_64FC1, rot[1]) };
  CvMat wimat[2] = { cvMat(3, 3, CV_64FC1, wi[0]), cvMat(3, 3, CV_64FC1, wi[1]) };

  bc = viam_bank_caldata(bank);

  /* get number of images */
  for(i=0; i<2/*stereo implies two cameras*/; i++) {
    cc = viam_image_caldata(viam_bank_image(vh, bank, i));
    if (i == 0) { nimages = cc->data->nimages; continue; }

    if (nimages != cc->data->nimages) {
      viam_seterrno(vh, Eviamlib_BAD_IMAGE_COUNT);
      return;
    }
  }

  /* rectify each pair of image separately */
  for(j=0;j<nimages;j++) {

    for(i=0; i<2/*stereo implies two cameras*/; i++) {
      /* compute IE products for image j */
      cc = viam_image_caldata(viam_bank_image(vh, bank, i));

      cvSetData(&_i, cc->c.intrinsic, sizeof(double [3]));
      cvSetData(&rvec, cc->data->rot + j, sizeof(CvPoint3D64f));
      cvRodrigues2(&rvec, &rotmat[i], NULL);
      e[0] = rot[i][0]; e[1] = rot[i][1]; e[2] = rot[i][2];
      e[4] = rot[i][3]; e[5] = rot[i][4]; e[6] = rot[i][5];
      e[8] = rot[i][6]; e[9] = rot[i][7]; e[10]= rot[i][8];
      e[7] = cc->data->trans[j].y;
      e[3] = cc->data->trans[j].x;
      e[11]= cc->data->trans[j].z;
      cvMatMul(&_i, &emat, &pmat[i]);
    }

    /* estimate rectification matrices */
    viam_getrectification(p[0], p[1],
			  0, 0, cc->c.width-1, cc->c.height-1,
			  &base, &pbase, w[0], w[1], rot[0], rot[1], ir);

    /* reproject calibration rig and measure error */
    cvInvert(&wmat[0], &wimat[0], CV_SVD);
    cvInvert(&wmat[1], &wimat[1], CV_SVD);

    err = 0.;
    for(k=0; k<nimages*bc->data->rows*bc->data->cols; k++) {
      CvPoint2D32f f[2];

      for(i=0; i<2/*stereo implies two cameras*/; i++) {
	double x,y,w;

	cc = viam_image_caldata(viam_bank_image(vh, bank, i));
	undistort(cc->data->points[k], &f[i], cc->c.intrinsic, cc->c.distortion);

	x = wi[i][0]*f[i].x + wi[i][1]*f[i].y + wi[i][2];
	y = wi[i][3]*f[i].x + wi[i][4]*f[i].y + wi[i][5];
	w = wi[i][6]*f[i].x + wi[i][7]*f[i].y + wi[i][8];

	f[i].x = x/w;
	f[i].y = y/w;
      }
      err += fabs(f[0].y - f[1].y);
    }
    err /= k;
    printf("rectification: %f\n", err);

    if (j==0 || err < *precision) {
      /* store best rectification so far */
      *precision = err;
      bc->b.baseline = base;
      bc->b.pbaseline = pbase;

      for(i=0; i<2/*stereo implies two cameras*/; i++) {
	cc = viam_image_caldata(viam_bank_image(vh, bank, i));

	memcpy(cc->c.intrirect, &ir, sizeof(cc->c.intrirect));
	memcpy(cc->c.rectification, w[i], sizeof(cc->c.rectification));
	memcpy(cc->c.rotation, rot[i], sizeof(cc->c.rotation));
      }
    }
  }
}


/* --- viam_getrectification ----------------------------------------- */

/* Computes the virtual PP1 and PP2 projection matrices of the virtual
 * cameras, so that the virtual cameras are exactly aligned.
 *
 * Cette fonction a t crite  l'origine par Pierrick, vers 1993. Elle
 * a t revue et trafique par Patricia et Hassan, et finalement Sid a
 * nettoy le tout, en supprimant notamment des paramtres en entre qui
 * permettaient de spcifier quelle rectification on voulait. Dsormais,
 * la fct dtermine les matrices des rectification de manire  ce que la
 * disparit soit nulle  l'infini. La fonction a finalement t intgre
 *  la lib StereoCorrelation par Simon en 12/98.
 *
 * \param[in] P1	Projection matrix (intrinsic * extrinsic, 3x4)
 * \param[in] P2	Projection matrix (intrinsic * extrinsic, 3x4)
 * \param[in] imin	Desired smallest column number
 * \param[in] jmin	Desired smallest line number
 * \param[in] imax	Desired highest column number
 * \param[in] jmax	Desired highest line number
 * \param[out] baseline		Baseline in meters
 * \param[out] pbaseline	Baseline in pixels
 * \param[out] W1	Rectification matrix (3x3)
 * \param[out] W2	Rectification matrix (3x3)
 * \param[out] R1	Rotation matrix from original to rectified frame
 * \param[out] R2	Rotation matrix from original to rectified frame
 * \param[out] intrinsic	Rectified intrinsic matrix
 */

static void
viam_getrectification(double P1[12], double P2[12],
		      int imin, int jmin, int imax, int jmax,
		      double *baseline, double *pbaseline,
		      double W1[9], double W2[9], double R1[9], double R2[9],
		      double intrinsinc[9])
{
  double C1[3], A1[3], H1[3], V1[3], T1[3], G1[3];
  double C2[3], A2[3], H2[3], V2[3], T2[3], G2[3];
  double A[3], G[3], T[3], H[3], V[3];
  double C12[3];
  double u0, ku, v0, kv;
  int i;

  /* first get the C A H V representation */
  viam_get_CAHV(P1, C1, A1, H1, V1);
  viam_get_CAHV(P2, C2, A2, H2, V2);

  /* AGT is the rotation from the world reference frame to the rectified
   * camera frame. The same rotation is used for left and right camera */
  viam_get_rectAGT(C1, A1, H1, V1, C2, A2, H2, V2, A, G, T);

  /* Calculate new scale factors and image center */
  viam_get_rectcoeff(C1,A1,H1,V1, C2,A2,H2,V2, A,G,T,
		     imin, jmin, imax, jmax, &u0, &ku, &v0, &kv);

  /* Construct new warping matrices W1 and W2 */
  /* V = kv*T + v0*A	H = ku*G + u0*A */
  for(i=0; i<3; i++) {
    H[i] = ku*G[i] + u0*A[i];
    V[i] = kv*T[i] + v0*A[i];
  }

  viam_warping_coeffs(A1, H1, V1, A, H, V, W1);
  viam_warping_coeffs(A2, H2, V2, A, H, V, W2);

  /* Construct rotation matrix to original camera frame */
#define dot_product(v1, v2)			\
  (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2])

  viam_get_AGT(A1, H1, V1, G1, T1);
  R1[0]=dot_product(G1, G); R1[1]=dot_product(G1, T); R1[2]=dot_product(G1, A);
  R1[3]=dot_product(T1, G); R1[4]=dot_product(T1, T); R1[5]=dot_product(T1, A);
  R1[6]=dot_product(A1, G); R1[7]=dot_product(A1, T); R1[8]=dot_product(A1, A);

  viam_get_AGT(A2, H2, V2, G2, T2);
  R2[0]=dot_product(G2, G); R2[1]=dot_product(G2, T); R2[2]=dot_product(G2, A);
  R2[3]=dot_product(T2, G); R2[4]=dot_product(T2, T); R2[5]=dot_product(T2, A);
  R2[6]=dot_product(A2, G); R2[7]=dot_product(A2, T); R2[8]=dot_product(A2, A);

#undef dot_product

  /* The baseline of the system */
  for(i=0; i<3; i++) C12[i] = C2[i] - C1[i];
  *baseline = sqrt(C12[0]*C12[0] + C12[1]*C12[1] + C12[2]*C12[2]);
  *pbaseline = *baseline * ku;

  /* rectified intrinsinc matrix */
  intrinsinc[0] = ku; intrinsinc[1] = 0.0; intrinsinc[2] = u0;
  intrinsinc[3] = 0.0; intrinsinc[4] = kv; intrinsinc[5] = v0;
  intrinsinc[6] = 0.0; intrinsinc[7] = 0.0; intrinsinc[8] = 1.0;

  printf ("Baseline : %f\n", *baseline);
  printf ("i0:%.2f j0: %.2f ki:%.2f kj:%.2f\n", u0, v0, ku, kv);
}


/* --- viam_get_CAHV ------------------------------------------------- */

/* Calculate CAHV from a given projection matrix P
 *
 * \param[in] P		Projection matrix 3x4.
 * \param[out] C	Optical center (3x1), solution of P.(x y z 1)t = 0.
 * \param[out] A	Optical axis (3x1) (normal)
 * \param[out] H	X vector of camera frame (3x1).
 * \param[out] V	Y vector of camera frame (3x1).
 */

static void
viam_get_CAHV(const double P[12],
	      double C[3], double A[3], double H[3], double V[3])
{
  double R[9], iR[9], T[3];
  CvMat Rmat = cvMat(3, 3, CV_64FC1, R);
  CvMat iRmat = cvMat(3, 3, CV_64FC1, iR);
  CvMat Tmat = cvMat(3, 1, CV_64FC1, T);
  CvMat Cmat = cvMat(3, 1, CV_64FC1, C);

  H[0] = P[0]; H[1] = P[1]; H[2] = P[2];
  V[0] = P[4]; V[1] = P[5]; V[2] = P[6];
  A[0] = P[8]; A[1] = P[9]; A[2] = P[10];

  R[0] = P[0]; R[1] = P[1]; R[2] = P[2];
  R[3] = P[4]; R[4] = P[5]; R[5] = P[6];
  R[6] = P[8]; R[7] = P[9]; R[8] = P[10];

  T[0] = -P[3]; T[1] = -P[7]; T[2] = -P[11];
  cvInvert(&Rmat, &iRmat, CV_SVD);
  cvMatMul(&iRmat, &Tmat, &Cmat);
}


/* --- viam_get_AGT -------------------------------------------------- */

static void
viam_get_AGT(const double A1[3], const double H1[3], const double V1[3],
	     double G[3], double T[3])
{
  int i;
  double cx, cy;
  double tnorm=0.0, gnorm=0.0;

  cx = H1[0]*A1[0] + H1[1]*A1[1] + H1[2]*A1[2];
  cy = V1[0]*A1[0] + V1[1]*A1[1] + V1[2]*A1[2];

  for (i=0;i<3;i++) {
    G[i] = H1[i] - cx*A1[i];
    T[i] = V1[i] - cy*A1[i];
    gnorm += G[i]*G[i];
    tnorm += T[i]*T[i];
  }
  gnorm = sqrt(gnorm);
  tnorm = sqrt(tnorm);

  for (i=0;i<3;i++) {
    G[i] /= gnorm;
    T[i] /= tnorm;
  }
}


/* --- viam_get_rectAGT ---------------------------------------------- */

static void
viam_get_rectAGT(const double C1[3], const double A1[3], const double H1[3],
		 const double V1[3], const double C2[3], const double A2[3],
		 const double H2[3], const double V2[3],
		 double A[3], double G[3], double T[3])
{
  double k;
  int i;

  /* vector joining the optical centers -> new G */
  for(i=0; i<3; i++) G[i] = C2[i] - C1[i];
  k = sqrt(G[0]*G[0] + G[1]*G[1] + G[2]*G[2]);
  G[0] /= k;
  G[1] /= k;
  G[2] /= k;

  /* new A, normal to G */
  for(i=0; i<3; i++) A[i] = (A1[i] + A2[i])/2.;
  k = A[0]*G[0] + A[1]*G[1] + A[2]*G[2];
  for(i=0; i<3; i++) A[i] = A[i] - k*G[i];
  k = sqrt(A[0]*A[0] + A[1]*A[1] + A[2]*A[2]);
  A[0] /= k;
  A[1] /= k;
  A[2] /= k;

  printf (">>> new A: angle A-A1 : %.2f, angle A-A2 %.2f\n",
	  acos(A[0]*A1[0] +A[1]*A1[1] + A[2]*A1[2]),
	  acos(A[0]*A2[0] +A[1]*A2[1] + A[2]*A2[2]) );

  /* new T = A x G */
  T[0] = A[1]*G[2] - A[2]*G[1];
  T[1] = A[2]*G[0] - A[0]*G[2];
  T[2] = A[0]*G[1] - A[1]*G[0];
}


/* --- viam_get_rectcoeff -------------------------------------------- */

/*
 * Visualisation "optimale" des images rectifiees:
 *	- Calcul de D : M = C1 + alpha * D
 *                                 Dte porteuse d'un pixel ds camera1
 *	- Calcul de K : K = Di.T / Di.A
 *	- Determination de i0' et ki' (ident. pour les 2 cameras)
 *	- Determination de j01', j02', kj'
 */

static void
viam_get_rectcoeff(const double C1[3], const double A1[3], const double H1[3],
		   const double V1[3], const double C2[3], const double A2[3],
		   const double H2[3], const double V2[3],
		   double A[3], double G[3], double T[3],
		   int umin, int vmin, int umax, int vmax,
		   double *u0, double *ku, double *v0, double *kv)
{
#define mini(a,b) ((a) < (b) ? (a) : (b))
#define maxi(a,b) ((a) > (b) ? (a) : (b))

  /* Ki coefficient, image 1, top/bottom left/right */
  double Ku1_tl, Ku1_tr, Ku1_bl, Ku1_br;
  double Ku2_tl, Ku2_tr, Ku2_bl, Ku2_br;
  double Kv1_tl, Kv1_tr, Kv1_bl, Kv1_br;
  double Kv2_tl, Kv2_tr, Kv2_bl, Kv2_br;

  double K1, K2, K3, K4;

  int width  = umax - umin + 1;
  int height = vmax - vmin + 1;

  /* coord. of extremal pixels in the rectified frame */
  viam_get_K(umin, vmin, V1, H1, A1, T, G, A, &Ku1_tl, &Kv1_tl);
  viam_get_K(umax, vmin, V1, H1, A1, T, G, A, &Ku1_tr, &Kv1_tr);
  viam_get_K(umin, vmax, V1, H1, A1, T, G, A, &Ku1_bl, &Kv1_bl);
  viam_get_K(umax, vmax, V1, H1, A1, T, G, A, &Ku1_br, &Kv1_br);

  viam_get_K(umin, vmin, V2, H2, A2, T, G, A, &Ku2_tl, &Kv2_tl);
  viam_get_K(umax, vmin, V2, H2, A2, T, G, A, &Ku2_tr, &Kv2_tr);
  viam_get_K(umin, vmax, V2, H2, A2, T, G, A, &Ku2_bl, &Kv2_bl);
  viam_get_K(umax, vmax, V2, H2, A2, T, G, A, &Ku2_br, &Kv2_br);

  /* line coefficients: v0' and kv' */
  K1 = mini(Kv1_tl, Kv1_tr);               /* K3 = smallest common line */
  K2 = mini(Kv2_tl, Kv2_tr);
  K3 = maxi(K1, K2);

  K1 = maxi(Kv1_bl, Kv1_br);               /* K4 = greatest common line */
  K2 = maxi(Kv2_bl, Kv2_br);
  K4 = mini(K1, K2);

  *v0 = height / (1 - (K4 / K3));
  *kv = height / (K4 - K3);

  /* column coefficients: u0' and ku' */
  K1 = mini(Ku1_tl, Ku1_bl);
  K2 = mini(Ku2_tl, Ku2_bl);
  K3 = maxi(K1, K2);

  K1 = maxi(Ku1_tr, Ku1_br);
  K2 = maxi(Ku2_tr, Ku2_br);
  K4 = mini(K1, K2);

  *u0 = width / (1 - (K4 / K3));
  *ku = width / (K4 - K3);

#undef mini
#undef maxi
}


/* --- viam_get_K ---------------------------------------------------- */

/** Calculates Kx and Ky scaling factor for a given point (u,v) */

static void
viam_get_K(int u, int v,
	   const double Vi[3], const double Hi[3], const double Ai[3],
	   const double T[3], const double G[3], const double A[3],
	   double *Kx, double *Ky)
{
  double t1[3], t2[3], D[3];
  double deno;

  /* t1 = u*Ai - Hi */
  t1[0] = u*Ai[0] - Hi[0];
  t1[1] = u*Ai[1] - Hi[1];
  t1[2] = u*Ai[2] - Hi[2];

  /* t2 = v*Ai - Vi */
  t2[0] = v*Ai[0] - Vi[0];
  t2[1] = v*Ai[1] - Vi[1];
  t2[2] = v*Ai[2] - Vi[2];

  /* D = (u*Ai - Hi)x(v*Ai - Vi) where i=no.cam */
  D[0] = t1[1]*t2[2] - t1[2]*t2[1];
  D[1] = t1[2]*t2[0] - t1[0]*t2[2];
  D[2] = t1[0]*t2[1] - t1[1]*t2[0];

  /* Calcul de Ki et Kj */
  deno = D[0]*A[0] + D[1]*A[1] + D[2]*A[2];
  *Kx = (D[0]*G[0] + D[1]*G[1] + D[2]*G[2]) / deno;
  *Ky = (D[0]*T[0] + D[1]*T[1] + D[2]*T[2]) / deno;
}


/* --- viam_warping_coeffs ------------------------------------------- */

static void
viam_warping_coeffs(double A1[3], double H1[3], double V1[3],
		    double A[3], double H[3], double V[3],
		    double W[9])
{
  double r1[3], r2[3], r3[3];

#define cross_product(v, v1, v2)		\
  do {						\
    v[0] = v1[1]*v2[2] - v1[2]*v2[1];		\
    v[1] = v1[2]*v2[0] - v1[0]*v2[2];		\
    v[2] = v1[0]*v2[1] - v1[1]*v2[0];		\
  } while(0)

#define dot_product(v1, v2)			\
  (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2])

  /* direction (u1,v1) */
  cross_product(r1, V, A);
  cross_product(r2, A, H);
  cross_product(r3, H, V);

  /*  M[0] = r1[0]; M[1] = r2[0]; M[2] = r3[0];
   *  M[3] = r1[1]; M[4] = r2[1]; M[5] = r3[1];
   *  M[6] = r1[2]; M[7] = r2[2]; M[8] = r3[2];
   *
   *  M2[0] = H1[0]; M2[1] = H1[1]; M2[2] = H1[2];
   *  M2[3] = V1[0]; M2[4] = V1[1]; M2[5] = V1[2];
   *  M2[6] = A1[0]; M2[7] = A1[1]; M2[8] = A1[2];
   *
   * W = M2 * M
   */

  W[0]=dot_product(H1, r1); W[1]=dot_product(H1, r2); W[2]=dot_product(H1, r3);
  W[3]=dot_product(V1, r1); W[4]=dot_product(V1, r2); W[5]=dot_product(V1, r3);
  W[6]=dot_product(A1, r1); W[7]=dot_product(A1, r2); W[8]=dot_product(A1, r3);

  W[0]/= W[8]; W[1]/= W[8]; W[2]/= W[8];
  W[3]/= W[8]; W[4]/= W[8]; W[5]/= W[8];
  W[6]/= W[8]; W[7]/= W[8]; W[8] = 1.;
#undef dot_product
#undef cross_product
}


/* --- fillquad ------------------------------------------------------------ */

/** Fill quadrahedral, using Bresenham algorithm.
 *
 * \param[in] img	Image
 * \param[in] v		Points of the quadrahedral
 */

static void
fillquad(IplImage *img, CvPoint2D32f v[4])
{
#define FIXEDPT_SHIFT	16

  struct {
    int idx, di;
    int x, dx, ye;
  } edge[2];

  unsigned short pxsize = ((img->depth & 0xff)>>3)*img->nChannels;
  unsigned char *line, *pixel;
  int left = 0, right = 1;
  int xmin, xmax, ymin, ymax, imin;
  int i, c, y,  x1, x2;
  int edges = 4;

  /* compute extremums */

  xmin = xmax = v[0].x;
  ymin = ymax = v[0].y;
  imin = 0;
  for(i=1; i<4; i++) {
    if (v[i].y < ymin) {
      ymin = v[i].y;
      imin = i;
    }
    ymax = ymax > v[i].y ? ymax : v[i].y;
    xmax = xmax > v[i].x ? xmax : v[i].x;
    xmin = xmin < v[i].x ? xmin : v[i].x;
  }

  if (xmax < 0 || ymax < 0 || xmin >= img->width || ymin >= img->height)
    return;

  if (ymax > img->height - 1)
    ymax = img->height - 1;

  edge[0].idx = edge[1].idx = imin;
  edge[0].ye  = edge[1].ye  = ymin;
  edge[0].di = 1;
  edge[1].di = 3;

  /* loop over edges */
  line = (unsigned char *)img->imageData + img->widthStep * ymin;
  y = ymin;

  for(y = ymin; y <= ymax; y++, line += img->widthStep) {
    /* update current scanned edges */
    for(i=0; i<2; i++) {
      if (y >= edge[i].ye) {
	int idx = edge[i].idx, di = edge[i].di;
	int xs = 0, xe, ye, ty = 0;

	for(;;) {
	  ty = v[idx].y;
	  if (ty > y || edges == 0) break;
	  xs = v[idx].x;
	  idx += di;
	  idx -= ((idx < 4) - 1) & 4;   /* idx -= idx >= npts ? npts : 0 */
	  edges--;
	}

	ye = ty;
	xs <<= FIXEDPT_SHIFT;
	xe = (int)(v[idx].x) << FIXEDPT_SHIFT;

	/* no more edges */
	if (y >= ye) return;

	edge[i].ye = ye;
	edge[i].dx = ((xe - xs)*2 + (ye - y)) / (2 * (ye - y));
	edge[i].x = xs;
	edge[i].idx = idx;
      }
    }

    /* scanlines are form left to right */
    if (edge[left].x > edge[right].x) {
      left ^= 1;
      right ^= 1;
    }

    /* fill one scanline */
    x1 = edge[left].x;
    x2 = edge[right].x;
    if (y >= 0) {
      int xx1 = x1 >> FIXEDPT_SHIFT;
      int xx2 = x2 >> FIXEDPT_SHIFT;

      if (xx2 >= 0 && xx1 < img->width) {
	if (xx1 < 0) xx1 = 0;
	if (xx2 >= img->width - 1) xx2 = img->width - 1;

	pixel = line + xx1 * pxsize;
	for(i = xx2-xx1+1; i > 0; i--, pixel += pxsize)
	  for(c=0; c<img->nChannels; c++)
	    switch (img->depth & 0xff) {
	      case 8:	pixel[c] *= 0.6; break;
	      case 16:	((unsigned short *)pixel)[c] *= 0.6; break;
	    }
      }
    }

    x1 += edge[left].dx;
    x2 += edge[right].dx;

    edge[left].x = x1;
    edge[right].x = x2;
  }

#undef FIXEDPT_SHIFT
}


/* --- undistort ----------------------------------------------------- */

/** Compensate distortion for a single point
 *
 * \param[in] o		Original point
 * \param[out] u	Point with distortion corrected
 * \param[in] a		Intrinsic matrix
 * \param[in] coeff	Distortion coefficients (RT)
 */

static void
undistort(CvPoint2D32f o, CvPoint2D32f *u, const double a[9], const double coeff[5])
{
  int i;

  /* compensate distortion iteratively */
  double x = (o.x - a[2])/a[0], y = (o.y - a[5])/a[4], x0 = x, y0 = y;
  for(i = 0; i < 5; i++) {
    double r2 = x*x + y*y;
    double icdist = 1./(1 + (coeff[0] + (coeff[1] + coeff[2]*r2)*r2)*r2);
    double delta_x = 2*coeff[3]*x*y + coeff[4]*(r2 + 2*x*x);
    double delta_y = coeff[3]*(r2 + 2*y*y) + 2*coeff[4]*x*y;
    x = (x0 - delta_x)*icdist;
    y = (y0 - delta_y)*icdist;
  }

  u->x = x*a[0] + a[2];
  u->y = y*a[4] + a[5];
}


/* --- viam_initrectifmap -------------------------------------------- */

/** \brief Compute rectification map
 *
 * \param[in] A		Initrinsic matrix
 * \param[in] distort	Distortion coefficients (RT)
 * \param[in] rectif	Warping matrix
 * \param[out] mapx	Array of x-coordinates of the map
 * \param[out] mapy	Array of y-coordinates of the map
 */

void
viam_initrectifmap(const double a[9], const double distort[5],
		   const double rectif[9],
		   float *mapx, float *mapy, int width, int height)
{
  int i, j;
  float u, v, z;
  float u0, v0, fx, fy, _fx, _fy, k1, k2, k3, p1, p2;

  u0 = a[2]; v0 = a[5];
  fx = a[0]; fy = a[4];
  _fx = 1.f/fx; _fy = 1.f/fy;
  k1 = distort[0]; k2 = distort[1]; k3 = distort[2];
  p1 = distort[3]; p2 = distort[4];

  for(j=0; j<height; j++) for(i=0; i<width; i++,mapx++,mapy++) {
    z = 1./(rectif[6]*i + rectif[7]*j + rectif[8]);

    u = (rectif[0]*i + rectif[1]*j + rectif[2])*z;
    v = (rectif[3]*i + rectif[4]*j + rectif[5])*z;

    {
      float y = (v - v0)*_fy;
      float y2 = y*y;
      float _2p1y = 2*p1*y;
      float _3p1y2 = 3*p1*y2;
      float p2y2 = p2*y2;

      float x = (u - u0)*_fx;
      float x2 = x*x;
      float r2 = x2 + y2;
      float d = 1 + (k1 + (k2 + k3*r2)*r2)*r2;
      float _u = fx*(x*(d + _2p1y) + p2y2 + (3*p2)*x2) + u0;
      float _v = fy*(y*(d + (2*p2)*x) + _3p1y2 + p1*x2) + v0;

      *mapx = _u;
      *mapy = _v;
    }
  }
}

#endif /* VIAM_OPENCV_SUPPORT */
