/*
 * Copyright (c) 2007-2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *                                      Anthony Mallet on Wed Jan 24 2007
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include "opencv2/core/core_c.h"
#endif


/* --- local data ---------------------------------------------------- */

typedef enum viam_filterdst_t {
  VIAM_FDST_SOURCE,
  VIAM_FDST_SWAP_A,
  VIAM_FDST_SWAP_B,
  VIAM_FDST_FINAL
} viam_filterdst_t;

struct viam_filter {
  viam_filter_t next, prec;	/**< image filter list */
  viam_filter_t fnext;		/**< global filter list */

  int configured;		/**< true if configured */
  char *name;			/**< local name */
  viam_filtertype_t type;	/**< filter type */
  viam_image_ref image;		/**< image on which the filter applies */

  viam_filtermethod_t method;	/**< filter application method */
  viam_filterautomode_t autoarg;/**< values computation (man, auto, ...) */
  union viam_filterargs_obj args;/**< current filter values */

  viam_filterdst_t dst;		/**< destination buffer (swap or final) */
  IplImage *source;		/**< pointer to source image */
  IplImage result;		/**< image header of the filtered source */
  viam_geometry_t geom;		/**< resulting geometry */

  void *data;			/**< filter-specific data */
};

static const struct viam_filterfunc_t {
  const char *name;
  int (*setup)(viam_handle_t vh, viam_image_ref camera,
	viam_filterargs_obj args, viam_geometry_t *geom, void **data);
  int (*filter)(viam_handle_t vh,  viam_image_ref camera,
	const struct _IplImage *source, struct _IplImage *result,
	viam_filterargs_obj args, void *data);
  int (*autoarg)(viam_handle_t vh, viam_image_ref camera,
	const IplImage *source, viam_filterautomode_t *aumode,
	viam_filterargs_obj args, void *data);
} filter_funcs[] = {
#ifdef VIAM_OPENCV_SUPPORT
  [VIAM_FILTER_SUBSAMPLE] = {
    "subsampling", viam_geo_geosetup, viam_geo_subsample, NULL
  },
#endif
  [VIAM_FILTER_SCALE] = {
    "scaling", viam_geo_geosetup, viam_geo_scale, NULL
  },
  [VIAM_FILTER_BRIGHTNESS] = {
    "brightness", NULL, NULL, NULL
  },
  [VIAM_FILTER_EXPOSURE] = {
    "exposure", NULL, NULL, NULL
  },
  [VIAM_FILTER_GAIN] = {
    "gain", NULL, NULL, NULL
  },
  [VIAM_FILTER_GAMMA] = {
    "gamma", NULL, NULL, NULL
  },
  [VIAM_FILTER_SHUTTER] = {
    "shutter", NULL, NULL, NULL
  },
  [VIAM_FILTER_ZOOM] = {
    "zoom", NULL, NULL, NULL
  },
  [VIAM_FILTER_FOCUS] = {
    "focus", NULL, NULL, NULL
  },
  [VIAM_FILTER_FORMAT] = {
    "format", viam_format_geosetup, viam_cvtformat, NULL
  },
#ifdef VIAM_OPENCV_SUPPORT
  [VIAM_FILTER_DISTORT] = {
    "distortion compensation", viam_geo_distortsetup, viam_geo_remap, NULL
  },
  [VIAM_FILTER_DISTORT_NN] = {
    "distortion compensation", viam_geo_distortsetup, viam_geo_remap_nn, NULL
  },
  [VIAM_FILTER_RECTIFY] = {
    "rectification", viam_geo_rectifysetup, viam_geo_remap, NULL
  },
  [VIAM_FILTER_RECTIFY_NN] = {
    "rectification", viam_geo_rectifysetup, viam_geo_remap_nn, NULL
  },
#endif /* VIAM_OPENCV_SUPPORT */
  [VIAM_FILTER_WHITE_BALANCE] = {
    "white balance", viam_color_geosetup, viam_awb, viam_awb_auto
  },
#ifdef VIAM_OPENCV_SUPPORT
  [VIAM_FILTER_LOG] = {
    "laplacian of gaussian", NULL, viam_improc_log, NULL
  }
#endif /* VIAM_OPENCV_SUPPORT */
};

static viam_filter_t	viam_filter_create(viam_handle_t vh,
				const char *name);
static const char *	viam_filter_methoddescr(viam_filtermethod_t method);
static const char *	viam_filter_autodescr(viam_filterautomode_t aumode);


/* --- viam_filter_inithw -------------------------------------------- */

/** \brief Iterate over filters and initialize hardware parameters.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object.
 * \return 0 on success.
 */

int
viam_filter_inithw(viam_handle_t vh, viam_image_ref image)
{
  viam_camera_t camera = viam_image_camera(image);
  viam_filter_t f;
  int s;

  /* iterate over all filters */
  for(f = viam_image_flist(image); f; f=f->next) {

    /* set initial value of hardware filters */
    if (f->method == VIAM_FILTER_HARDWARE) {
      switch(f->autoarg) {
	case VIAM_FILTER_HARDWARE_AUTO:
	case VIAM_FILTER_HARDWARE_ONE_PUSH:
	  /* nothing to do */
	  break;

	default:
	  s = viam_sethwfeaturevalue(vh, camera, f->type, &f->args);
	  if (s) return s;
	  break;
      }
    }
  }

  return 0;
}


/* --- viam_filter_image --------------------------------------------- */

/** \brief Iterate over filters and apply them on image.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object.
 * \param[out] final	Filtered image.
 *
 * \return 0 on success.
 */

int
viam_filter_image(viam_handle_t vh, viam_image_ref image, IplImage *final)
{
  viam_camera_t camera = viam_image_camera(image);
  viam_filter_t f;
  int copied;
  int s;

  /* iterate over all filters */
  for(f = viam_image_flist(image), copied = 0; f; f=f->next) {

    /* compute new filter values if needed */
    if (f->autoarg == VIAM_FILTER_SOFTWARE_AUTO ||
	f->autoarg == VIAM_FILTER_SOFTWARE_ONE_PUSH) {
      s = filter_funcs[f->type].autoarg(
	vh, image, f->source, &f->autoarg, &f->args, f->data);
      if (s) return s;

      if (f->autoarg == VIAM_FILTER_MANUAL) {
	warnx("one push %s filter on image %s stopped",
	      filter_funcs[f->type].name, viam_image_name(image));
      }
    }

    /* apply value, depending on the method */
    switch(f->method) {
      case VIAM_FILTER_SOFTWARE:
	switch(f->dst) {
	  case VIAM_FDST_SOURCE: break;

	  case VIAM_FDST_SWAP_A:
	  case VIAM_FDST_SWAP_B:
	    cvSetData(&f->result, viam_swap(f->dst), CV_AUTOSTEP);
	    break;

	  case VIAM_FDST_FINAL:
	    cvSetData(&f->result, final->imageData, final->widthStep);
	    break;
	}

	s = filter_funcs[f->type].filter(
	  vh, image, f->source, &f->result, &f->args, f->data);
	if (s) return s;
	copied = 1;
	break;

      case VIAM_FILTER_HARDWARE:
	cvSetData(&f->result, f->source->imageData, f->source->widthStep);
	switch(f->autoarg) {
          case VIAM_FILTER_MANUAL:
	  case VIAM_FILTER_HARDWARE_AUTO:
	  case VIAM_FILTER_HARDWARE_ONE_PUSH:
	    /* nothing to do */
	    break;

	  default:
	    s = viam_sethwfeaturevalue(vh, camera, f->type, &f->args);
	    if (s) return s;
	    break;
	}
	break;

      case VIAM_FILTER_INVALID: break;
      case VIAM_FILTER_OFF: break;
    }
  }

  /* special case: if no filter has copied the image yet, copy it
   * manually. Check that source and destination have the same widthStep to
   * make a single memcpy in this case */
  if (!copied) {
    IplImage *source = viam_camera_raw(camera);
    char *sptr, *dptr;
    size_t w;
    int i;

    if (source->widthStep == final->widthStep)
      memcpy(final->imageData, source->imageData, final->imageSize);
    else {
      sptr = source->imageData;
      dptr = final->imageData;
      w = (source->widthStep < final->widthStep) ?
	source->widthStep : final->widthStep;

      for(i = final->height; i>0; i--) {
        memcpy(dptr, sptr, w);
        sptr+= source->widthStep;
        dptr+= final->widthStep;
      }
    }
  }

  return 0;
}


/* --- viam_filter_push ---------------------------------------------- */

/** \brief Install filter on an image.
 *
 * The camera must not be configured.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object.
 * \param[in] filter	Filter object.
 * \param[in] method	Filter method.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

int
viam_filter_push(viam_handle_t vh,
		 viam_image_ref image, viam_filter_t filter,
		 viam_filtermethod_t method, viam_filterautomode_t aumode)
{
  viam_camera_t camera = viam_image_camera(image);
  viam_filter_t prec;
  int s;

  /* check filter is not used yet */
  if (filter->image) return viam_seterrno(vh, Eviamlib_CONFIGURED);

  /* check mode has only one bit set (i.e. exactly one mode) */
  if (method & (method-1)) {
    warnx("bad method %d", method);
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }
  if (aumode & (aumode-1)) {
    warnx("bad auto mode %d", aumode);
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  /* find previous filter */
  for(prec = viam_image_flist(image); prec && prec->next; prec = prec->next)
    /* empty body */;

  /* push filter into the list */
  filter->next = NULL;
  filter->prec = prec;
  filter->image = image;
  filter->method = method;
  filter->autoarg = aumode;
  filter->dst = VIAM_FDST_FINAL; /* since we are the last filter */
  filter->source = prec ? &prec->result : viam_camera_raw(camera);
  filter->geom.width = filter->source->width;
  filter->geom.height = filter->source->height;
  filter->geom.scalex = prec ? prec->geom.scalex : 1.;
  filter->geom.scaley = prec ? prec->geom.scaley : 1.;
  filter->geom.offx = prec ? prec->geom.offx : 0.;
  filter->geom.offy = prec ? prec->geom.offy : 0.;
  filter->geom.depth = filter->source->depth;
  filter->geom.channels = filter->source->nChannels;
  filter->data = NULL;
  filter->configured = 0;

  /* configure geometry and modes */
  s = viam_filter_configure(vh, image, filter);
  if (s) return viam_errno(vh);

  /* insert new filter at tail */
  if (prec) prec->next = filter; else viam_image_setflist(image, filter);

  warnx("added %s filter to image %s",
	filter_funcs[filter->type].name, viam_image_name(image));
  return 0;
}


/* --- viam_filter_update -------------------------------------------- */

/** \brief Update filter parameters.
 *
 * This function cannot change filter geometry
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter object.
 * \param[in] method	New filter method.
 * \param[in] aumode	New auto mode.
 * \param[in] args	New values.
 *
 * \return 0 on success.
 */

static int
viam_filter_update(viam_handle_t vh, viam_filter_t filter,
		   viam_filtermethod_t method, viam_filterautomode_t aumode,
		   viam_filterargs_obj args)
{
  viam_camera_t camera = viam_image_camera(filter->image);
  struct viam_filter prec;
  int s;

  if (!camera)
    return viam_seterrno(vh, Eviamlib_NOT_CONFIGURED);

  /* check mode has only one bit set (i.e. exactly one mode) */
  if (method & (method-1)) {
    warnx("bad method %d", method);
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }
  if (aumode & (aumode-1)) {
    warnx("bad auto mode %d", aumode);
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  /* backup previous settings in case of error */
  prec = *filter;
  filter->method = method;
  filter->autoarg = aumode;
  filter->args = *args;

  /* (re)configure modes (geometry won't be allowed to change) */
  s = viam_filter_configure(vh, filter->image, filter);
  if (s) { *filter = prec; return viam_errno(vh); }

  return 0;
}


/* --- viam_filter_configure ----------------------------------------- */

/** \brief (Re)configure images geometry and filter mode.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object.
 * \param[in] f		Filter list.
 *
 * \return 0 on success.
 */

int
viam_filter_configure(viam_handle_t vh, viam_image_ref image, viam_filter_t f)
{
  viam_camera_t camera = viam_image_camera(image);
  unsigned int methods, aumodes;
  viam_geometry_t ngeo, *pgeo;
  union viam_filterargs_obj nargs;
  viam_filterdst_t dst;
  IplImage *final;
  int s, count;

  /* get initial geometry */
  if (f && f->prec) pgeo = &f->prec->geom; else pgeo = NULL;

  /* iterate on the list */
  for(; f; pgeo = &f->geom, f = f->next) {
    /* if camera is not yet configured, stop here */
    if (!viam_camera_setup(camera)) continue;

    /* check filter mode */
    s = viam_queryfeature(vh, camera, f->type, &methods, &aumodes);
    if (s) return s;

    if (!(methods & f->method)) {
      warnx("camera %s has no %s %s", viam_camera_name(camera),
	    viam_filter_methoddescr(f->method), filter_funcs[f->type].name);
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
    }
    if (!(aumodes & f->autoarg)) {
      warnx("camera %s has no %s %s", viam_camera_name(camera),
	    viam_filter_autodescr(f->autoarg), filter_funcs[f->type].name);
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
    }

    /* check and compute geometry */
    ngeo.width = f->source->width;
    ngeo.height = f->source->height;
    ngeo.scalex = pgeo ? pgeo->scalex : 1.;
    ngeo.scaley = pgeo ? pgeo->scaley : 1.;
    ngeo.offx = pgeo ? pgeo->offx : 0.;
    ngeo.offy = pgeo ? pgeo->offy : 0.;
    ngeo.depth = f->source->depth;
    ngeo.channels = f->source->nChannels;

    nargs = f->args;

    if (filter_funcs[f->type].setup) {
      s = filter_funcs[f->type].setup(vh, image, &nargs, &ngeo, &f->data);
      if (s) {
	warnx("%s filter on image %s has invalid parameters",
	      filter_funcs[f->type].name, viam_image_name(image));
	return s;
      }
    }

    if (f->configured) {
      /* already configured filters cannot change their geometry */
      if (ngeo.width != f->geom.width || ngeo.height != f->geom.height ||
	  ngeo.depth != f->geom.depth || ngeo.channels != f->geom.channels) {
	warnx(
	  "%s filter cannot change geometry", filter_funcs[f->type].name);
	return viam_seterrno(vh, Eviamlib_CONFIGURED);
      }
    } else {
      /* initialize intermediate image header */
      cvInitImageHeader(&f->result, cvSize(ngeo.width, ngeo.height),
			ngeo.depth, ngeo.channels, IPL_ORIGIN_TL, 4);
    }

    /* configure hw filter mode and any static value */
    s = viam_sethwfeaturemode(vh, camera, f->type, f->method, f->autoarg);
    if (s) return s;
    if (f->method == VIAM_FILTER_HARDWARE && f->autoarg == VIAM_FILTER_MANUAL) {
      s = viam_sethwfeaturevalue(vh, camera, f->type, &f->args);
      if (s) return s;
    }

    /* done */
    f->geom = ngeo;
    f->args = nargs;
    f->configured = 1;

    if (!f->next) {
      warnx("filters on image %s produce %dx%d images, %d bpp, %d channel%s",
	    viam_image_name(image), ngeo.width, ngeo.height, ngeo.depth,
	    ngeo.channels, ngeo.channels>1?"s":"");
    }
  }

  /* configure swap and images destinations for all filters */
  count = 0;
  for(f = viam_image_flist(image); f; f = f->next) {
    if (f->method == VIAM_FILTER_SOFTWARE) count++;
    if (!f->next) final = &f->result;
  }
  dst = (count%2) ? VIAM_FDST_FINAL : VIAM_FDST_SWAP_A;

  for(f = viam_image_flist(image); f; f = f->next) {
    f->dst = dst;
    switch(f->dst) {
      case VIAM_FDST_SWAP_A:
      case VIAM_FDST_SWAP_B:
	viam_setswap(f->dst, f->result.imageSize);
	break;

      case VIAM_FDST_FINAL:
	if (f->result.imageSize > final->imageSize) {
	  f->dst = VIAM_FDST_SWAP_B;
	  viam_setswap(f->dst, f->result.imageSize);
	}
	break;

      case VIAM_FDST_SOURCE: break;
    }

    if (f->method == VIAM_FILTER_SOFTWARE)
      dst = (dst == VIAM_FDST_FINAL) ? VIAM_FDST_SWAP_A : VIAM_FDST_FINAL;
  }

  return 0;
}


/* --- viam_filter_geo ----------------------------------------------- */

/** \brief Get filtered image geometry.
 *
 * \param[in] vh	Library handle.
 * \param[in] f		Filter list.
 * \param[out] geom	Resulting image geometry after applying f
 *                      filters.
 *
 * \return 0 on success.
 */

int
viam_filter_geo(viam_handle_t vh, viam_filter_t f, viam_geometry_t *geom)
{
  for(; f && f->next; f = f->next)
    /* empty body */;
  if (!f) return viam_seterrno(vh, Eviamlib_IMAGE_GEOMETRY);

  *geom = f->geom;
  return 0;
}


/* --- viam_filter_queryswfeature ------------------------------------ */

/** \brief Return software filter capabilities.
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter type.
 * \param[out] methods	Or'ed combination of capabilities.
 * \param[out] aumodes	Or'ed combination of auto modes.
 *
 * \return 0 on success.
 */

int
viam_filter_queryswfeature(viam_handle_t vh, viam_filtertype_t filter,
			   unsigned int *methods, unsigned int *aumodes)
{
  if (methods) *methods = 0;
  if (aumodes) *aumodes = 0;

  if (filter_funcs[filter].filter) {
    if (methods) *methods |= VIAM_FILTER_SOFTWARE;
    if (aumodes) *aumodes |= VIAM_FILTER_MANUAL;
  }

  if (filter_funcs[filter].autoarg && aumodes) {
    *aumodes |= VIAM_FILTER_SOFTWARE_AUTO;
    *aumodes |= VIAM_FILTER_SOFTWARE_ONE_PUSH;
  }

  return 0;
}


/* --- viam_filter_format_create ------------------------------------- */

/** \brief Create a filter modifying format of images.
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Name to give to the new filter.
 * \param[in] type	Filter type: specific format algorithm.
 * \param[in] format	Image format selection.
 */

viam_filter_t
viam_filter_format_create(viam_handle_t vh, const char *name,
			  viam_filtertype_t type, viam_filterformat_t format)
{
  viam_filter_t f = viam_filter_create(vh, name);
  if (!f) return NULL;

  switch(type) {
    case VIAM_FILTER_FORMAT:
      break;

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f->type = type;
  f->args.format = format;
  return f;
}


/* --- viam_filter_format_update ------------------------------------- */

/** \brief Update parameters of a format filter.
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter to be updated
 * \param[in] method	New filter method
 * \param[in] aumode	New filter auto mode
 * \param[in] format	New format value
 *
 * \return reconfigured filter.
 */

int
viam_filter_format_update(viam_handle_t vh, viam_filter_t filter,
			  viam_filtermethod_t method,
			  viam_filterautomode_t aumode, viam_filterformat_t format)
{
  union viam_filterargs_obj args;

  switch(filter->type) {
    case VIAM_FILTER_FORMAT:
      break;

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  args.format = format;
  return viam_filter_update(vh, filter, method, aumode, &args);
}


/* --- viam_filter_geometry_create ----------------------------------- */

/** \brief Create a geometry filter.
 *
 * Filter will crop and scale image according to new size
 * specification. If sizes are floats between 0 and 1, they are
 * interpreted as a percentage of the original dimension. Otherwise they
 * are rounded to the nearest integer.
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Name to give to the new filter.
 * \param[in] type	Filter type: specific geometry algorithm.
 * \param[in] left	Left corner.
 * \param[in] top	Top corner.
 * \param[in] bottom	Right corner.
 * \param[in] right	Bottom corner.
 * \param[in] towidth	Desired window width.
 * \param[in] toheight	Desired window height.
 */

viam_filter_t
viam_filter_geometry_create(viam_handle_t vh,
			    const char *name, viam_filtertype_t type,
			    double left, double top,
			    double right, double bottom,
			    double towidth, double toheight)
{
  viam_filter_t f;

  switch(type) {
    case VIAM_FILTER_SCALE:
#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_SUBSAMPLE:
    case VIAM_FILTER_DISTORT:
    case VIAM_FILTER_DISTORT_NN:
    case VIAM_FILTER_RECTIFY:
    case VIAM_FILTER_RECTIFY_NN:
#endif /* VIAM_OPENCV_SUPPORT */
      break;

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f = viam_filter_create(vh, name);
  if (!f) return NULL;

  f->type = type;
  f->args.left = left;
  f->args.top = top;
  f->args.right = right;
  f->args.bottom = bottom;
  f->args.towidth = towidth;
  f->args.toheight = toheight;
  return f;
}


/* --- viam_filter_geometry_update ----------------------------------- */

/** \brief Update parameters of a geometry filter
 * (subsample, rectify, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter to be updated
 * \param[in] method	New filter method
 * \param[in] aumode	New filter auto mode
 * \param[in] left	Left corner.
 * \param[in] top	Top corner.
 * \param[in] bottom	Right corner.
 * \param[in] right	Bottom corner.
 * \param[in] towidth	Desired window width.
 * \param[in] toheight	Desired window height.
 *
 * \return reconfigured filter.
 */

int
viam_filter_geometry_update(viam_handle_t vh, viam_filter_t filter,
			    viam_filtermethod_t method,
			    viam_filterautomode_t aumode,
			    double left, double top,
			    double right, double bottom,
			    double towidth, double toheight)
{
  union viam_filterargs_obj args;

  switch(filter->type) {
    case VIAM_FILTER_SCALE:
#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_SUBSAMPLE:
    case VIAM_FILTER_DISTORT:
    case VIAM_FILTER_DISTORT_NN:
    case VIAM_FILTER_RECTIFY:
    case VIAM_FILTER_RECTIFY_NN:
#endif /* VIAM_OPENCV_SUPPORT */
      break;

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  args.left = left;
  args.top = top;
  args.right = right;
  args.bottom = bottom;
  args.towidth = towidth;
  args.toheight = toheight;
  return viam_filter_update(vh, filter, method, aumode, &args);
}


/* --- viam_filter_luminance_create ---------------------------------- */

/** \brief Create a luminance filter (brightness, exposure, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Filter name.
 * \param[in] type	Filter type: specific luminance algorithm.
 * \param[in] parameter	Value
 *
 * \return initialized filter.
 */

viam_filter_t
viam_filter_luminance_create(viam_handle_t vh, const char *name,
			     viam_filtertype_t type, viam_valuekind_t vkind,
			     double parameter)
{
  viam_filter_t f;

  switch(type) {
    case VIAM_FILTER_EXPOSURE:
    case VIAM_FILTER_BRIGHTNESS:
    case VIAM_FILTER_SHARPNESS:
    case VIAM_FILTER_GAIN:
    case VIAM_FILTER_SHUTTER:
      break;

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f = viam_filter_create(vh, name);
  if (!f) return NULL;

  f->type = type;
  f->args.vkind = vkind;
  f->args.value = parameter;
  return f;
}


/* --- viam_filter_luminance_update ---------------------------------- */

/** \brief Update parameters of a luminance filter
 * (brightness, exposure, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter to be updated
 * \param[in] method	New filter method
 * \param[in] aumode	New filter auto mode
 * \param[in] parameter	New filter value
 *
 * \return reconfigured filter.
 */

int
viam_filter_luminance_update(viam_handle_t vh, viam_filter_t filter,
			     viam_filtermethod_t method,
			     viam_filterautomode_t aumode,
			     viam_valuekind_t vkind, double parameter)
{
  union viam_filterargs_obj args;

  switch(filter->type) {
    case VIAM_FILTER_EXPOSURE:
    case VIAM_FILTER_BRIGHTNESS:
    case VIAM_FILTER_SHARPNESS:
    case VIAM_FILTER_GAIN:
    case VIAM_FILTER_SHUTTER:
      break;

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  args.vkind = vkind;
  args.value = parameter;
  return viam_filter_update(vh, filter, method, aumode, &args);
}


/* --- viam_filter_color_create -------------------------------------- */

/** \brief Create a color filter (white balance, gamma, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Filter name.
 * \param[in] type	Filter type: specific color algorithm.
 * \param[in] parameter	Value
 *
 * \return initialized filter.
 */

viam_filter_t
viam_filter_color_create(viam_handle_t vh, const char *name,
			 viam_filtertype_t type, viam_valuekind_t vkind,
			 double parameter)
{
  viam_filter_t f;

  switch(type) {
    case VIAM_FILTER_WHITE_BALANCE:
    case VIAM_FILTER_GAMMA:
      break;

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f = viam_filter_create(vh, name);
  if (!f) return NULL;

  f->type = type;
  f->args.vkind = vkind;
  if (type == VIAM_FILTER_WHITE_BALANCE) {
    f->args.blue = ((int)parameter)/100.;
    f->args.red = parameter - (int)parameter;
  } else {
    f->args.value = parameter;
  }

  return f;
}


/* --- viam_filter_color_update -------------------------------------- */

/** \brief Update parameters of a color filter
 * (white balance, gamma, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter to be updated
 * \param[in] method	New filter method
 * \param[in] aumode	New filter auto mode
 * \param[in] parameter	New filter value
 *
 * \return reconfigured filter.
 */

int
viam_filter_color_update(viam_handle_t vh, viam_filter_t filter,
			 viam_filtermethod_t method,
			 viam_filterautomode_t aumode, viam_valuekind_t vkind,
			 double parameter)
{
  union viam_filterargs_obj args;

  switch(filter->type) {
    case VIAM_FILTER_WHITE_BALANCE:
    case VIAM_FILTER_GAMMA:
      break;

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  args.vkind = vkind;
  if (filter->type == VIAM_FILTER_WHITE_BALANCE) {
    args.blue = ((int)parameter)/100.;
    args.red = parameter - (int)parameter;
  } else {
    args.value = parameter;
  }

  return viam_filter_update(vh, filter, method, aumode, &args);
}


/* --- viam_filter_improc_create ------------------------------------- */

/** \brief Create an image processing filter (LoG, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Filter name.
 * \param[in] type	Filter type: specific color algorithm.
 * \param[in] parameter	Value
 *
 * \return initialized filter.
 */

viam_filter_t
viam_filter_improc_create(viam_handle_t vh, const char *name,
			  viam_filtertype_t type, viam_valuekind_t vkind,
			  double parameter)
{
  viam_filter_t f;

  switch(type) {
#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_LOG:
      break;
#endif /* VIAM_OPENCV_SUPPORT */

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f = viam_filter_create(vh, name);
  if (!f) return NULL;

  f->type = type;
  f->args.vkind = vkind;
  f->args.value = parameter;

  return f;
}


/* --- viam_filter_improc_update ------------------------------------- */

/** \brief Update parameters of an image processing filter (LoG, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter to be updated
 * \param[in] method	New filter method
 * \param[in] aumode	New filter auto mode
 * \param[in] parameter	New filter value
 *
 * \return reconfigured filter.
 */

int
viam_filter_improc_update(viam_handle_t vh, viam_filter_t filter,
			  viam_filtermethod_t method,
			  viam_filterautomode_t aumode, viam_valuekind_t vkind,
			  double parameter)
{
  union viam_filterargs_obj args;

  switch(filter->type) {
#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_LOG:
      break;
#endif /* VIAM_OPENCV_SUPPORT */

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  args.vkind = vkind;
  args.value = parameter;
  return viam_filter_update(vh, filter, method, aumode, &args);
}


/* --- viam_filter_misc_create --------------------------------------------- */

/** \brief Create a miscelaneous filter (Zoom, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Filter name.
 * \param[in] type	Filter type: specific filter name.
 * \param[in] parameter	Value
 *
 * \return initialized filter.
 */

viam_filter_t
viam_filter_misc_create(viam_handle_t vh, const char *name,
			viam_filtertype_t type, viam_valuekind_t vkind,
			double parameter)
{
  viam_filter_t f;

  switch(type) {
    case VIAM_FILTER_ZOOM:
    case VIAM_FILTER_FOCUS:
    case VIAM_FILTER_PAN:
    case VIAM_FILTER_TILT:
      break;

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f = viam_filter_create(vh, name);
  if (!f) return NULL;

  f->type = type;
  f->args.vkind = vkind;
  f->args.value = parameter;

  return f;
}


/* --- viam_filter_misc_update --------------------------------------------- */

/** \brief Update parameters of a miscelaneous filter (Zoom, ...).
 *
 * \param[in] vh	Library handle.
 * \param[in] filter	Filter to be updated
 * \param[in] method	New filter method
 * \param[in] aumode	New filter auto mode
 * \param[in] parameter	New filter value
 *
 * \return reconfigured filter.
 */

int
viam_filter_misc_update(viam_handle_t vh, viam_filter_t filter,
			viam_filtermethod_t method,
			viam_filterautomode_t aumode, viam_valuekind_t vkind,
			double parameter)
{
  union viam_filterargs_obj args;

  switch(filter->type) {
    case VIAM_FILTER_ZOOM:
    case VIAM_FILTER_FOCUS:
    case VIAM_FILTER_PAN:
    case VIAM_FILTER_TILT:
      break;

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  args.vkind = vkind;
  args.value = parameter;
  return viam_filter_update(vh, filter, method, aumode, &args);
}


/* --- viam_filter_create -------------------------------------------- */

/** \brief Create a generic, empty filter structure.
 *
 * \param[in] vh	Library handle
 * \param[in] name	Filter name
 *
 * \return the empty structure.
 */

static viam_filter_t
viam_filter_create(viam_handle_t vh, const char *name)
{
  viam_filter_t f;

  f = malloc(sizeof(*f));
  if (f) f->name = strdup(name);

  if (!f || !f->name) {
    warnx("unable to allocate memory");
    if (f) free(f);
    viam_seterrno(vh, Eviamlib_MEMORY);
    return NULL;
  }

  f->next = f->prec = NULL;
  f->fnext = viam_config_filters(vh);
  f->configured = 0;
  f->type = VIAM_FILTER_NONE;
  f->image = NULL;
  f->method = VIAM_FILTER_INVALID;
  f->autoarg = VIAM_FILTER_AUTOINVALID;
  f->args = (union viam_filterargs_obj){{ 0 }};
  f->dst = VIAM_FDST_SOURCE;
  f->source = NULL;
  f->result = (IplImage){0};
  f->geom = (viam_geometry_t){0};
  f->data = NULL;
  viam_config_setfilters(vh, f);

  return f;
}


/* --- viam_filter_getbyname ----------------------------------------- */

/** \brief Get a filter by its local name
 *
 * \param[in] vh	Library handle.
 * \param[in] name	Local name of the filter.
 *
 * \return the image handle, or NULL if not found
 */

viam_filter_t
viam_filter_getbyname(viam_handle_t vh, const char *name)
{
  viam_filter_t f;

  for(f = viam_config_filters(vh); f; f = f->fnext)
    if (!strcmp(f->name, name)) return f;

  viam_seterrno(vh, Eviamlib_NO_SUCH_FILTER);
  return NULL;
}


/* --- viam_filter mutators and accessors ---------------------------- */

const char *
viam_filter_name(const viam_filter_t filter)
{
  return filter->name;
}

viam_filtertype_t
viam_filter_type(const viam_filter_t filter)
{
  return filter->type;
}

viam_filtermethod_t
viam_filter_method(const viam_filter_t filter)
{
  return filter->method;
}

viam_filterautomode_t
viam_filter_autoarg(const viam_filter_t filter)
{
  return filter->autoarg;
}


/* --- viam_filter_methoddescr --------------------------------------- */

static const char *
viam_filter_methoddescr(viam_filtermethod_t method)
{
  switch(method) {
    case VIAM_FILTER_INVALID:	return "invalid";
    case VIAM_FILTER_OFF:	return "hardware off";
    case VIAM_FILTER_HARDWARE:	return "hardware";
    case VIAM_FILTER_SOFTWARE:	return "software";
  }

  return "*null*";
}


/* --- viam_filter_autoescr ------------------------------------------ */

static const char *
viam_filter_autodescr(viam_filterautomode_t aumode)
{
  switch(aumode) {
    case VIAM_FILTER_AUTOINVALID:	return "invalid";
    case VIAM_FILTER_MANUAL:		return "manual";
    case VIAM_FILTER_HARDWARE_AUTO:	return "hardware automatic";
    case VIAM_FILTER_HARDWARE_ONE_PUSH:	return "hardware one push";
    case VIAM_FILTER_SOFTWARE_AUTO:	return "software automatic";
    case VIAM_FILTER_SOFTWARE_ONE_PUSH:	return "software one push";
  }

  return "*null*";
}
